"""Contains test cases for the downloaders.py module."""

import sys
import unittest
from pathlib import Path

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui import downloaders

YOUTUBEDL_OUTPUT_VIDEO = PATH.joinpath(
    "fixtures/extract_data_video_output.txt"
).read_text(encoding="utf-8")
YOUTUBEDL_OUTPUT_PLAYLIST = PATH.joinpath(
    "fixtures/extract_data_playlist_output.txt"
).read_text(encoding="utf-8")

OUTPUT_VIDEO_EXPECTED = [
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {
        "status": "Downloading",
        "path": "",
        "filename": "Among Us Momentos Divertidos-among-us-momentos-divertidos-2020-09-24-18-28-12-525257.f0",
        "extension": ".mp4",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "1.13MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.3%",
        "filesize": "1.13MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.6%",
        "filesize": "1.13MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "1.3%",
        "filesize": "1.13MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "2.7%",
        "filesize": "1.13MiB",
        "speed": "330.59KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "5.4%",
        "filesize": "1.13MiB",
        "speed": "192.02KiB/s",
        "eta": "00:05",
    },
    {
        "status": "Downloading",
        "percent": "10.9%",
        "filesize": "1.13MiB",
        "speed": "198.19KiB/s",
        "eta": "00:05",
    },
    {
        "status": "Downloading",
        "percent": "22.0%",
        "filesize": "1.13MiB",
        "speed": "211.94KiB/s",
        "eta": "00:04",
    },
    {
        "status": "Downloading",
        "percent": "41.6%",
        "filesize": "1.13MiB",
        "speed": "261.50KiB/s",
        "eta": "00:02",
    },
    {
        "status": "Downloading",
        "percent": "72.2%",
        "filesize": "1.13MiB",
        "speed": "314.87KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "1.13MiB",
        "speed": "370.95KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "1.13MiB",
    },
    {
        "status": "Downloading",
        "path": "",
        "filename": "Among Us Momentos Divertidos-among-us-momentos-divertidos-2020-09-24-18-28-12-525257.f3",
        "extension": ".m4a",
    },
    {
        "status": "Downloading",
        "percent": "0.2%",
        "filesize": "569.83KiB",
        "speed": "64.50KiB/s",
        "eta": "00:08",
    },
    {
        "status": "Downloading",
        "percent": "0.5%",
        "filesize": "569.83KiB",
        "speed": "193.50KiB/s",
        "eta": "00:02",
    },
    {
        "status": "Downloading",
        "percent": "1.2%",
        "filesize": "569.83KiB",
        "speed": "451.50KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "2.6%",
        "filesize": "569.83KiB",
        "speed": "967.50KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "5.4%",
        "filesize": "569.83KiB",
        "speed": "1.95MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "11.1%",
        "filesize": "569.83KiB",
        "speed": "672.93KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "22.3%",
        "filesize": "569.83KiB",
        "speed": "739.49KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "44.8%",
        "filesize": "569.83KiB",
        "speed": "859.19KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "89.7%",
        "filesize": "569.83KiB",
        "speed": "884.08KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "569.83KiB",
        "speed": "810.46KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "569.83KiB",
    },
    {
        "status": "Post Processing",
        "path": "",
        "filename": "Among Us Momentos Divertidos-among-us-momentos-divertidos-2020-09-24-18-28-12-525257",
        "extension": ".mp4",
    },
    {},
    {},
]
OUTPUT_PLAYLIST_EXPECTED = [
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {"status": "Downloading"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {"status": "Downloading", "playlist_index": "1", "playlist_size": "3"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {
        "status": "Downloading",
        "path": "",
        "filename": "USO EFICIENTE DE DATOS-uso-eficiente-datos-2021-04-21-08-42-41-766550.f2",
        "extension": ".mp4",
    },
    {
        "status": "Downloading",
        "percent": "0.0%",
        "filesize": "2.62MiB",
        "speed": "64.07KiB/s",
        "eta": "00:41",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "2.62MiB",
        "speed": "192.20KiB/s",
        "eta": "00:13",
    },
    {
        "status": "Downloading",
        "percent": "0.3%",
        "filesize": "2.62MiB",
        "speed": "448.46KiB/s",
        "eta": "00:05",
    },
    {
        "status": "Downloading",
        "percent": "0.6%",
        "filesize": "2.62MiB",
        "speed": "960.98KiB/s",
        "eta": "00:02",
    },
    {
        "status": "Downloading",
        "percent": "1.2%",
        "filesize": "2.62MiB",
        "speed": "1.94MiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "2.4%",
        "filesize": "2.62MiB",
        "speed": "672.02KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "4.7%",
        "filesize": "2.62MiB",
        "speed": "739.22KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "9.5%",
        "filesize": "2.62MiB",
        "speed": "1.06MiB/s",
        "eta": "00:02",
    },
    {
        "status": "Downloading",
        "percent": "19.1%",
        "filesize": "2.62MiB",
        "speed": "1.39MiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "38.2%",
        "filesize": "2.62MiB",
        "speed": "1.73MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "76.4%",
        "filesize": "2.62MiB",
        "speed": "2.13MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "2.62MiB",
        "speed": "2.46MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "2.62MiB",
    },
    {
        "status": "Downloading",
        "path": "",
        "filename": "USO EFICIENTE DE DATOS-uso-eficiente-datos-2021-04-21-08-42-41-766550.f5",
        "extension": ".m4a",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "1.28MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.2%",
        "filesize": "1.28MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.5%",
        "filesize": "1.28MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "1.1%",
        "filesize": "1.28MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "2.4%",
        "filesize": "1.28MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "4.8%",
        "filesize": "1.28MiB",
        "speed": "807.38KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "9.7%",
        "filesize": "1.28MiB",
        "speed": "813.19KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "19.4%",
        "filesize": "1.28MiB",
        "speed": "1020.19KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "38.9%",
        "filesize": "1.28MiB",
        "speed": "991.25KiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "77.8%",
        "filesize": "1.28MiB",
        "speed": "1.52MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "1.28MiB",
        "speed": "1.79MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "1.28MiB",
    },
    {
        "status": "Post Processing",
        "path": "",
        "filename": "USO EFICIENTE DE DATOS-uso-eficiente-datos-2021-04-21-08-42-41-766550",
        "extension": ".mp4",
    },
    {},
    {},
    {"status": "Downloading", "playlist_index": "2", "playlist_size": "3"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {
        "status": "Downloading",
        "path": "",
        "filename": "VIDEOTUTORIAL APKLIS-videotutorial-apklis-2021-04-21-08-32-02-115958.f2",
        "extension": ".mp4",
    },
    {
        "status": "Downloading",
        "percent": "0.0%",
        "filesize": "3.51MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "3.51MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.2%",
        "filesize": "3.51MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.4%",
        "filesize": "3.51MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.9%",
        "filesize": "3.51MiB",
        "speed": "1.94MiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "1.8%",
        "filesize": "3.51MiB",
        "speed": "671.81KiB/s",
        "eta": "00:05",
    },
    {
        "status": "Downloading",
        "percent": "3.5%",
        "filesize": "3.51MiB",
        "speed": "738.85KiB/s",
        "eta": "00:04",
    },
    {
        "status": "Downloading",
        "percent": "7.1%",
        "filesize": "3.51MiB",
        "speed": "960.07KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "14.2%",
        "filesize": "3.51MiB",
        "speed": "1.33MiB/s",
        "eta": "00:02",
    },
    {
        "status": "Downloading",
        "percent": "28.4%",
        "filesize": "3.51MiB",
        "speed": "1.64MiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "56.9%",
        "filesize": "3.51MiB",
        "speed": "1.85MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "3.51MiB",
        "speed": "1.74MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "3.51MiB",
    },
    {
        "status": "Downloading",
        "path": "",
        "filename": "VIDEOTUTORIAL APKLIS-videotutorial-apklis-2021-04-21-08-32-02-115958.f5",
        "extension": ".m4a",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "1.57MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.2%",
        "filesize": "1.57MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.4%",
        "filesize": "1.57MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.9%",
        "filesize": "1.57MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "1.9%",
        "filesize": "1.57MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "3.9%",
        "filesize": "1.57MiB",
        "speed": "1008.08KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "7.9%",
        "filesize": "1.57MiB",
        "speed": "812.77KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "15.9%",
        "filesize": "1.57MiB",
        "speed": "1020.26KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "31.8%",
        "filesize": "1.57MiB",
        "speed": "1.23MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "63.6%",
        "filesize": "1.57MiB",
        "speed": "1.78MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "1.57MiB",
        "speed": "2.39MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "1.57MiB",
    },
    {
        "status": "Post Processing",
        "path": "",
        "filename": "VIDEOTUTORIAL APKLIS-videotutorial-apklis-2021-04-21-08-32-02-115958",
        "extension": ".mp4",
    },
    {},
    {},
    {"status": "Downloading", "playlist_index": "3", "playlist_size": "3"},
    {"status": "Pre Processing"},
    {"status": "Pre Processing"},
    {
        "status": "Downloading",
        "path": "",
        "filename": "ENRED2 CORTAFUEGOS WIFI-enred2-cortafuegos-wifi-2021-04-14-09-49-03-477520.f2",
        "extension": ".mp4",
    },
    {
        "status": "Downloading",
        "percent": "0.0%",
        "filesize": "2.40MiB",
        "speed": "32.08KiB/s",
        "eta": "01:16",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "2.40MiB",
        "speed": "96.24KiB/s",
        "eta": "00:25",
    },
    {
        "status": "Downloading",
        "percent": "0.3%",
        "filesize": "2.40MiB",
        "speed": "224.56KiB/s",
        "eta": "00:10",
    },
    {
        "status": "Downloading",
        "percent": "0.6%",
        "filesize": "2.40MiB",
        "speed": "481.20KiB/s",
        "eta": "00:05",
    },
    {
        "status": "Downloading",
        "percent": "1.3%",
        "filesize": "2.40MiB",
        "speed": "661.86KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "2.6%",
        "filesize": "2.40MiB",
        "speed": "672.30KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "5.2%",
        "filesize": "2.40MiB",
        "speed": "625.53KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "10.4%",
        "filesize": "2.40MiB",
        "speed": "709.79KiB/s",
        "eta": "00:03",
    },
    {
        "status": "Downloading",
        "percent": "20.8%",
        "filesize": "2.40MiB",
        "speed": "1.10MiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "41.6%",
        "filesize": "2.40MiB",
        "speed": "1.45MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "83.3%",
        "filesize": "2.40MiB",
        "speed": "1.58MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "100.0%",
        "filesize": "2.40MiB",
        "speed": "1.71MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "speed": "",
        "eta": "",
        "percent": "100%",
        "filesize": "2.40MiB",
    },
    {
        "status": "Downloading",
        "path": "",
        "filename": "ENRED2 CORTAFUEGOS WIFI-enred2-cortafuegos-wifi-2021-04-14-09-49-03-477520.f5",
        "extension": ".m4a",
    },
    {
        "status": "Downloading",
        "percent": "0.1%",
        "filesize": "1.29MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.2%",
        "filesize": "1.29MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "0.5%",
        "filesize": "1.29MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "1.1%",
        "filesize": "1.29MiB",
        "speed": "Unknown",
        "eta": "ETA",
    },
    {
        "status": "Downloading",
        "percent": "2.3%",
        "filesize": "1.29MiB",
        "speed": "1.95MiB/s",
        "eta": "00:00",
    },
    {
        "status": "Downloading",
        "percent": "4.8%",
        "filesize": "1.29MiB",
        "speed": "672.94KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "9.6%",
        "filesize": "1.29MiB",
        "speed": "739.59KiB/s",
        "eta": "00:01",
    },
    {
        "status": "Downloading",
        "percent": "19.2%",
        "filesize": "1.29MiB",
        "speed": "1020.62KiB/s",
        "eta": "00:01",
    },
]


class TestDownloaders(unittest.TestCase):
    def setUp(self) -> None:
        self.expected_output = {
            "video": OUTPUT_VIDEO_EXPECTED,
            "playlist": OUTPUT_PLAYLIST_EXPECTED,
        }

    def process_data(self, output: str, expected: str) -> None:
        lines = [line for line in output.splitlines() if line]

        STEPS = len(lines) // 10

        for step in range(STEPS):
            for num, line in enumerate(lines[step * 10 : (step + 1) * 10]):
                data = downloaders.extract_data(line)
                # print(data)
                self.assertDictEqual(
                    self.expected_output[expected][num + (step * 10)], data
                )

    def test_extract_data_video(self):
        self.process_data(YOUTUBEDL_OUTPUT_VIDEO, "video")

    def test_extract_data_playlist(self):
        self.process_data(YOUTUBEDL_OUTPUT_PLAYLIST, "playlist")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



import unittest

import wx


class WidgetTestCase(unittest.TestCase):
    """
    A testcase that will create an app and frame for various widget test
    modules to use. They can inherit from this class to save some work. This
    is also good for test cases that just need to have an application object
    created.
    """

    def setUp(self):
        self.app = wx.App()
        wx.Log.SetActiveTarget(wx.LogStderr())
        self.frame = wx.Frame(None, title="WTC: " + self.__class__.__name__)
        # self.frame.Show()
        # self.frame.PostSizeEvent()

    def tearDown(self):
        def _cleanup():
            for tlw in wx.GetTopLevelWindows():  # type: ignore
                if tlw:
                    if isinstance(tlw, wx.Dialog) and tlw.IsModal():
                        tlw.EndModal(0)
                        wx.CallAfter(tlw.Destroy)
                    else:
                        tlw.Close(force=True)
            wx.WakeUpIdle()

        timer = wx.PyTimer(_cleanup)
        timer.Start(100)
        self.app.MainLoop()
        del self.app



"""Contains test cases for the logmanager.py module."""

import sys
import unittest
from pathlib import Path
from unittest import mock

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.logmanager import LogManager


class TestLogManager(unittest.TestCase):
    def setUp(self) -> None:
        self.config_path: str = str(PATH.joinpath("fixtures"))

    def test_init(self):
        log_mng = LogManager(self.config_path, True)
        self.assertEqual(
            log_mng.log_file,
            str(Path(self.config_path) / Path(LogManager.LOG_FILENAME)),
        )

    @mock.patch("youtube_dl_gui.logmanager.LogManager", autospec=True)
    def test_log(self, mock_logmanager):
        opt_mng = mock_logmanager.return_value
        opt_mng.log(data="Logging from tests")
        opt_mng.log.assert_called_once()



"""Contains test cases for the optionsmanager.py module."""

import sys
import unittest
from pathlib import Path
from unittest import mock

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.optionsmanager import OptionsManager


class TestOptionsManager(unittest.TestCase):
    def setUp(self) -> None:
        self.config_path: str = str(PATH.joinpath("fixtures"))

    def test_init(self):
        opt_mng = OptionsManager(self.config_path)
        self.assertEqual(
            opt_mng.settings_file,
            str(Path(self.config_path) / Path(OptionsManager.SETTINGS_FILENAME)),
        )

    @mock.patch("youtube_dl_gui.optionsmanager.OptionsManager", autospec=True)
    def test_save_to_file(self, mock_optionsmanager):
        opt_mng = mock_optionsmanager.return_value
        opt_mng.save_to_file()
        opt_mng.save_to_file.assert_called_once()


def main():
    unittest.main()


if __name__ == "__main__":
    main()



# type: ignore[misc]
"""Custom widgets for yt-dlg"""
from __future__ import annotations

from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import wx
import wx.lib.masked as masked

from .darktheme import DARK_BACKGROUND_COLOUR, DARK_FOREGROUND_COLOUR, dark_mode
from .utils import IS_WINDOWS

if TYPE_CHECKING:
    from .downloadmanager import DownloadItem
    from .mainframe import MainFrame
    from .optionsframe import OptionsFrame

_: Callable[[str], str] = wx.GetTranslation


def crt_command_event(event: wx.PyEventBinder, event_id: int = 0) -> wx.CommandEvent:
    """Shortcut to create command events."""
    return wx.CommandEvent(event.typeId, event_id)


# noinspection PyUnresolvedReferences,PyPep8Naming
class ListBoxWithHeaders(wx.ListBox):
    """Custom ListBox object that supports 'headers'.

    Attributes:
        NAME (str): Default name for the name argument of the __init__.

        TEXT_PREFIX (str): Text to add before normal items in order to
            distinguish them (normal items) from headers.

        EVENTS (list): List with events to overwrite to avoid header selection.

    """

    NAME = "listBoxWithHeaders"

    TEXT_PREFIX = "    "

    EVENTS: list[wx.PyEventBinder] = [
        wx.EVT_LEFT_DOWN,
        wx.EVT_LEFT_DCLICK,
        wx.EVT_RIGHT_DOWN,
        wx.EVT_RIGHT_DCLICK,
        wx.EVT_MIDDLE_DOWN,
        wx.EVT_MIDDLE_DCLICK,
    ]

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        choices=[],
        style=0,
        validator=wx.DefaultValidator,
        name=NAME,
    ) -> None:
        super().__init__(parent, id, pos, size, choices, style, validator, name)
        self.__headers: set[str] = set()

        # Ignore all key events i'm bored to handle the header selection
        self.Bind(wx.EVT_KEY_DOWN, lambda event: None)

        # Make sure that a header is never selected
        self.Bind(wx.EVT_LISTBOX, self._on_listbox)
        for _event in self.EVENTS:
            self.Bind(_event, self._disable_header_selection)

        # Append the items in our own way in order to add the TEXT_PREFIX
        self.AppendItems(choices)

    def _disable_header_selection(self, event) -> None:
        """Stop event propagation if the selected item is a header."""
        row = self.HitTest(event.GetPosition())
        event_skip = True

        if row != wx.NOT_FOUND and self.GetString(row) in self.__headers:
            event_skip = False

        event.Skip(event_skip)

    def _on_listbox(self, event) -> None:
        """Make sure no header is selected."""
        if event.GetString() in self.__headers:
            self.Deselect(event.GetSelection())
        event.Skip()

    def _add_prefix(self, string: str) -> str:
        return self.TEXT_PREFIX + string

    def _remove_prefix(self, string: str) -> str:
        if string[: len(self.TEXT_PREFIX)] == self.TEXT_PREFIX:
            return string[len(self.TEXT_PREFIX) :]
        return string

    # wx.ListBox methods

    def FindString(self, string: str, **kwargs) -> int:
        index = super().FindString(string, **kwargs)

        if index == wx.NOT_FOUND:
            # This time try with prefix
            index = super().FindString(self._add_prefix(string), **kwargs)

        return index

    def GetStringSelection(self) -> str:
        return self._remove_prefix(super().GetStringSelection())

    def GetString(self, index: int) -> str:
        if index < 0 or index >= self.GetCount():
            # Return empty string based on the wx.ListBox docs
            # for some reason parent GetString does not handle
            # invalid indices
            return ""

        return self._remove_prefix(super().GetString(index))

    def InsertItems(self, items: list[str], pos: int) -> None:
        items = [self._add_prefix(item) for item in items]
        super().InsertItems(items, pos)

    def SetSelection(self, index: int) -> None:
        if self.GetString(index) in self.__headers:
            self.Deselect(self.GetSelection())
        else:
            super().SetSelection(index)

    def SetString(self, index: int, string: str) -> None:
        old_string = self.GetString(index)

        if old_string in self.__headers and string != old_string:
            self.__headers.remove(old_string)
            self.__headers.add(string)

        super().SetString(index, string)

    def SetStringSelection(self, string: str) -> bool:
        if string in self.__headers:
            return False

        self.SetSelection(
            self.FindString(
                string,
            )
        )
        return True

    # wx.ItemContainer methods

    def AppendItems(self, strings: list[str], with_prefix: bool = True) -> None:
        if with_prefix:
            strings = [self._add_prefix(string) for string in strings]

        super().AppendItems(strings)

    def Clear(self) -> None:
        self.__headers.clear()
        super().Clear()

    def Delete(self, index: int) -> None:
        string: str = self.GetString(index)

        if string in self.__headers:
            self.__headers.remove(string)

        super().Delete(index)

    # Extra methods

    def add_header(self, header_string: str) -> int:
        self.__headers.add(header_string)
        return super().Append(header_string)

    def add_item(
        self,
        item: str,
        with_prefix: bool = True,
        clientData: dict[str, str] | None = None,
    ) -> None:
        if with_prefix:
            item = self._add_prefix(item)

        super().Append(item, clientData)

    def add_items(self, items: list[str], with_prefix: bool = True) -> None:
        for item in items:
            self.add_item(item, with_prefix)


# noinspection PyPep8Naming
class ListBoxComboPopup(wx.ComboPopup):
    """ListBoxWithHeaders as a popup"""

    def __init__(
        self, parent: wx.ComboCtrl | None = None, darkmode: bool = False
    ) -> None:
        super().__init__()
        self.__parent = parent
        self.__listbox: ListBoxWithHeaders | None = None
        self.__dark_mode: bool = darkmode
        self.value = -1

    def _on_motion(self, event) -> None:
        row: int = self.__listbox.HitTest(event.GetPosition())

        if row != wx.NOT_FOUND:
            self.__listbox.SetSelection(row)

            self.curitem = row if self.__listbox.IsSelected(row) else wx.NOT_FOUND

    # noinspection PyUnusedLocal
    def _on_left_down(self, event) -> None:
        self.value = self.curitem
        if self.value >= 0:
            self.Dismiss()

    # wx.ComboPopup methods

    # noinspection PyAttributeOutsideInit
    def Init(self) -> None:
        self.value = self.curitem = -1

    def Create(self, parent: wx.ComboCtrl, **kwargs) -> bool:
        # Create components
        self.__listbox = ListBoxWithHeaders(parent, style=wx.LB_SINGLE)

        if self.__dark_mode:
            self.__listbox.SetBackgroundColour(DARK_BACKGROUND_COLOUR)
            self.__listbox.SetForegroundColour(DARK_FOREGROUND_COLOUR)

        self.__listbox.Bind(wx.EVT_MOTION, self._on_motion)
        self.__listbox.Bind(wx.EVT_LEFT_DOWN, self._on_left_down)

        return True

    def GetControl(self) -> ListBoxWithHeaders | None:
        return self.__listbox

    def AddItem(
        self,
        item: str,
        with_prefix: bool = True,
        clientData: dict[str, str] | None = None,
    ) -> None:
        self.__listbox.add_item(item, with_prefix, clientData)

    def AddItems(self, items: list[str], with_prefix: bool = True) -> None:
        self.__listbox.add_items(items, with_prefix)

    def GetStringValue(self) -> str:
        return self.__listbox.GetString(self.value)

    def GetSelection(self) -> int:
        return self.value

    def SetSelection(self, index: int) -> None:
        self.__listbox.SetSelection(index)

        if self.__listbox.IsSelected(index):
            self.value = index
            self.__parent.SetValue(self.GetStringValue())

    def SetStringSelection(self, string: str) -> None:
        index: int = self.__listbox.FindString(
            string,
        )
        self.__listbox.SetSelection(index)

        if index != wx.NOT_FOUND and self.__listbox.GetSelection() == index:
            self.value = index
            self.SetSelection(self.value)

    def Clear(self) -> None:
        self.__parent.SetValue("")
        self.__listbox.Clear()

    def IsListEmpty(self) -> bool:
        return self.__listbox.GetCount() == 0

    def OnDismiss(self) -> None:
        if self.value < 0:
            self.value = 0
            self.__listbox.SetSelection(self.value)


# noinspection PyPep8Naming
class ExtComboBox(wx.ComboBox):
    def __init__(self, parent, max_items=-1, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        assert max_items > 0 or max_items == -1
        self.max_items = max_items

    def Append(self, new_value):
        if self.FindString(new_value) == wx.NOT_FOUND:
            super().Append(new_value)

            if self.max_items != -1 and self.GetCount() > self.max_items:
                self.SetItems(self.GetStrings()[1:])

    def SetValue(self, new_value):
        index = self.FindString(new_value)

        if index == wx.NOT_FOUND:
            self.Append(new_value)

        self.SetSelection(index)

    def LoadMultiple(self, items_list):
        for item in items_list:
            self.Append(item)


class DoubleStageButton(wx.Button):
    def __init__(self, parent, labels, bitmaps, bitmap_pos=wx.TOP, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        assert isinstance(labels, tuple) and isinstance(bitmaps, tuple)
        assert len(labels) == 2
        assert len(bitmaps) in [0, 2]

        self.labels = labels
        self.bitmaps = bitmaps
        self.bitmap_pos = bitmap_pos

        self._stage = 0
        self._set_layout()

    def _set_layout(self):
        self.SetLabel(self.labels[self._stage])

        if len(self.bitmaps):
            self.SetBitmap(self.bitmaps[self._stage], self.bitmap_pos)

    def change_stage(self):
        self._stage = 0 if self._stage else 1
        self._set_layout()

    def set_stage(self, new_stage):
        assert new_stage in [0, 1]

        self._stage = new_stage
        self._set_layout()


class MessageDialog(wx.Dialog):
    STYLE = (
        wx.DEFAULT_DIALOG_STYLE
        if IS_WINDOWS
        else wx.DEFAULT_DIALOG_STYLE | wx.MAXIMIZE_BOX
    )

    def __init__(self, parent, message, title, _dark_mode=False):
        super().__init__(parent, wx.ID_ANY, title, style=self.STYLE)
        self.parent = parent
        self.message = message
        self._dark_mode = _dark_mode

        # Create components
        self.panel = wx.Panel(self)

        self.buttons: dict[str, wx.Button] = {
            "yes": wx.Button(self.panel, wx.ID_YES, _("Yes")),
            "no": wx.Button(self.panel, wx.ID_NO, _("No")),
        }

        info_bmp = wx.ArtProvider.GetBitmap(wx.ART_INFORMATION, wx.ART_MESSAGE_BOX)
        info_icon = wx.StaticBitmap(self.panel, wx.ID_ANY, info_bmp)

        msg_text = wx.StaticText(self.panel, wx.ID_ANY, message)

        # Set sizers
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        message_sizer = wx.BoxSizer(wx.HORIZONTAL)
        message_sizer.Add(info_icon)
        message_sizer.AddSpacer(10)
        message_sizer.Add(msg_text, flag=wx.EXPAND | wx.TOP | wx.BOTTOM, border=5)

        vertical_sizer.Add(message_sizer, 1, wx.ALL, border=10)

        buttons_sizer = wx.StdDialogButtonSizer()

        for button in self.buttons.values():
            button.Bind(wx.EVT_BUTTON, self._on_close)
            buttons_sizer.AddButton(button)

        self.buttons["no"].SetDefault()

        buttons_sizer.Realize()

        vertical_sizer.Add(buttons_sizer, flag=wx.EXPAND | wx.ALL, border=10)

        self.panel.SetSizer(vertical_sizer)

        width, height = self.panel.GetBestSize()
        self.SetSize((width, int(height * 1.8)))

        # Set Dark Theme
        dark_mode(self.panel, self._dark_mode)

        self.Center()

    def _on_close(self, event):
        self.EndModal(event.GetEventObject().GetId())


class ButtonsChoiceDialog(wx.Dialog):

    STYLE = (
        wx.DEFAULT_DIALOG_STYLE
        if IS_WINDOWS
        else wx.DEFAULT_DIALOG_STYLE | wx.MAXIMIZE_BOX
    )

    def __init__(self, parent, choices, message, title, _dark_mode=False):
        super().__init__(parent, wx.ID_ANY, title, style=self.STYLE)
        self._dark_mode = _dark_mode

        self.buttons: dict[str, wx.Button] = {}

        # Create components
        self.panel = wx.Panel(self)

        info_bmp = wx.ArtProvider.GetBitmap(wx.ART_INFORMATION, wx.ART_MESSAGE_BOX)

        info_icon = wx.StaticBitmap(self.panel, wx.ID_ANY, info_bmp)
        msg_text = wx.StaticText(self.panel, wx.ID_ANY, message)

        self.buttons["cancel"] = wx.Button(self.panel, wx.ID_CANCEL, _("Cancel"))

        for index, label in enumerate(choices):
            key: str = str(index + 1)
            self.buttons[key] = wx.Button(self.panel, int(key), label)

        # Get the maximum button width & height
        max_width = max_height = -1

        for button in self.buttons.values():
            button_width, button_height = button.GetSize()

            if button_width > max_width:
                max_width = button_width

            if button_height > max_height:
                max_height = button_height

        max_width += 10

        # Set buttons width & bind events
        for button in self.buttons.values():
            if button != self.buttons["cancel"]:
                button.SetMinSize((max_width, max_height))
            else:
                # On Cancel button change only the height
                button.SetMinSize((-1, max_height))

            button.Bind(wx.EVT_BUTTON, self._on_close)

        # Set sizers
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        message_sizer = wx.BoxSizer(wx.HORIZONTAL)
        message_sizer.Add(info_icon)
        message_sizer.AddSpacer(10)
        message_sizer.Add(msg_text, flag=wx.EXPAND | wx.TOP | wx.BOTTOM, border=5)

        vertical_sizer.Add(message_sizer, 1, wx.ALL, border=10)

        buttons_sizer = wx.BoxSizer(wx.HORIZONTAL)

        for button in (self.buttons["1"], self.buttons["2"]):
            buttons_sizer.Add(button)
            buttons_sizer.AddSpacer(5)

        buttons_sizer.AddSpacer(1)
        buttons_sizer.Add(self.buttons["cancel"])
        vertical_sizer.Add(buttons_sizer, flag=wx.EXPAND | wx.ALL, border=10)

        self.panel.SetSizer(vertical_sizer)

        width, height = self.panel.GetBestSize()
        self.SetSize((width, int(height * 1.5)))

        # Set Dark Theme
        dark_mode(self.panel, self._dark_mode)

        self.Center()

    def _on_close(self, event):
        self.EndModal(event.GetEventObject().GetId())


class ClipDialog(wx.Dialog):

    FRAME_SIZE = (195, 170) if IS_WINDOWS else (350, 250)

    CHECK_OPTIONS = ("--external-downloader", "--external-downloader-args")

    def __init__(
        self,
        parent: MainFrame,
        download_item: DownloadItem,
        _dark_mode: bool = False,
    ):
        super().__init__(
            parent, wx.ID_ANY, title=_("Clip Multimedia"), style=wx.DEFAULT_DIALOG_STYLE
        )
        self.download_item = download_item
        clip_start, clip_end = self._get_timespans()

        self._dark_mode = _dark_mode

        # Create components
        self.panel = wx.Panel(self)
        sizer = wx.BoxSizer(wx.VERTICAL)
        h_time_box = wx.BoxSizer(wx.HORIZONTAL)

        start_label = wx.StaticText(self.panel, wx.ID_ANY, _("Clip start") + ":")
        h_time_box.Add(start_label, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        h_time_box.AddStretchSpacer(1)
        self.clip_start = masked.TimeCtrl(
            self.panel, wx.ID_ANY, value=clip_start, fmt24hr=True, name="startTime"
        )
        height = self.clip_start.GetSize().height
        spin1 = wx.SpinButton(
            self.panel, wx.ID_ANY, wx.DefaultPosition, (-1, height), wx.SP_VERTICAL
        )
        self.clip_start.BindSpinButton(spin1)
        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        hbox1.Add(self.clip_start, 0, wx.ALIGN_CENTRE)
        hbox1.Add(spin1, 0, wx.ALIGN_CENTRE)

        h_time_box.Add(hbox1, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(h_time_box, 0, wx.EXPAND | wx.ALL, 5)

        h_time_box = wx.BoxSizer(wx.HORIZONTAL)

        end_label = wx.StaticText(self.panel, wx.ID_ANY, _("Clip end") + ":")
        h_time_box.Add(end_label, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        h_time_box.AddStretchSpacer(1)
        spin2 = wx.SpinButton(
            self.panel, wx.ID_ANY, wx.DefaultPosition, (-1, height), wx.SP_VERTICAL
        )
        self.clip_end = masked.TimeCtrl(
            self.panel,
            wx.ID_ANY,
            value=clip_end,
            fmt24hr=True,
            name="endTime",
            spinButton=spin2,
        )
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        hbox2.Add(self.clip_end, 0, wx.ALIGN_CENTRE)
        hbox2.Add(spin2, 0, wx.ALIGN_CENTRE)

        h_time_box.Add(hbox2, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(h_time_box, 0, wx.EXPAND | wx.ALL, 5)

        line = wx.StaticLine(
            self.panel, wx.ID_ANY, size=(-1, -1), style=wx.LI_HORIZONTAL
        )
        sizer.Add(line, 0, wx.EXPAND | wx.TOP, 5)

        buttons_sizer = wx.StdDialogButtonSizer()

        btn = wx.Button(self.panel, wx.ID_OK)
        btn.Bind(wx.EVT_BUTTON, self._on_close)
        btn.SetDefault()

        buttons_sizer.AddButton(btn)

        btn = wx.Button(self.panel, wx.ID_CANCEL)
        buttons_sizer.AddButton(btn)
        buttons_sizer.Realize()

        sizer.Add(buttons_sizer, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 5)

        self.panel.SetSizer(sizer)
        # Set Dark Theme
        dark_mode(self.panel, self._dark_mode)

        self.SetSize(self.FRAME_SIZE)
        self.Center()
        # TODO: Make better decision
        self._clean_options()

    def _clean_options(self):
        """
        Clean the CHECK_OPTIONS from self.download_item.options

        """
        options = []

        for idx, opt in enumerate(self.download_item.options):
            if opt in self.CHECK_OPTIONS:
                try:
                    opt_arg = self.download_item.options[idx + 1]
                    if opt_arg == "ffmpeg" or "-ss" in opt_arg or "-to" in opt_arg:
                        self.download_item.options.pop(idx + 1)
                except IndexError:
                    pass

                continue

            options.append(opt)

        if options:
            self.download_item.options = options

    def _get_timespans(self) -> tuple[str, str]:
        """
        Get the TimeSpan if CHECK_OPTIONS[1] in self.download_item.options

        Returns:
            Tuple of strings with the clip_start and clip_end in format HH:MM:SS

        """
        external_downloader_args: str | None = None
        downloader_args: list[str] | None = None
        clip_start = clip_end = index = 0

        for idx, option in enumerate(self.download_item.options):
            if self.CHECK_OPTIONS[1] == option:
                try:
                    external_downloader_args = self.download_item.options[idx + 1]
                    index = idx
                except IndexError:
                    # No exist timespans
                    self.download_item.options.pop(idx)
                break

        if external_downloader_args:
            downloader_args = external_downloader_args.split()

        if downloader_args and len(downloader_args) == 4:
            # Clean quotes (simple/double)
            try:
                clip_start = int(downloader_args[1].strip("'\""))
                clip_end = int(downloader_args[-1].strip("'\""))
            except ValueError:
                self.download_item.options.pop(index + 1)
                self.download_item.options.pop(index)

        wx_clip_start = str(timedelta(seconds=clip_start))
        wx_clip_end = str(timedelta(seconds=clip_end))
        return wx_clip_start, wx_clip_end

    def _on_close(self, event):
        _clip_start = int(self.clip_start.GetValue(as_wxTimeSpan=True).GetSeconds())
        _clip_end = int(self.clip_end.GetValue(as_wxTimeSpan=True).GetSeconds())

        if _clip_start > _clip_end or _clip_start == _clip_end:
            wx.MessageBox(
                _("Invalid timespan"), _("Warning"), wx.OK | wx.ICON_EXCLAMATION
            )
            return

        self.EndModal(event.GetEventObject().GetId())


class ShutdownDialog(wx.Dialog):
    STYLE = (
        wx.DEFAULT_DIALOG_STYLE
        if IS_WINDOWS
        else wx.DEFAULT_DIALOG_STYLE | wx.MAXIMIZE_BOX
    )

    TIMER_INTERVAL = 1000  # milliseconds

    def __init__(self, parent, timeout, message, *args, **kwargs):
        super().__init__(parent, wx.ID_ANY, *args, style=self.STYLE, **kwargs)
        assert timeout > 0

        self.timeout = timeout
        self.message = message

        # Create components
        panel = wx.Panel(self)

        info_bmp = wx.ArtProvider.GetBitmap(wx.ART_INFORMATION, wx.ART_MESSAGE_BOX)
        info_icon = wx.StaticBitmap(panel, wx.ID_ANY, info_bmp)

        self.msg_text = msg_text = wx.StaticText(panel, wx.ID_ANY, self._get_message())
        ok_button = wx.Button(panel, wx.ID_OK, _("OK"))
        cancel_button = wx.Button(panel, wx.ID_CANCEL, _("Cancel"))

        # Set layout
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        message_sizer = wx.BoxSizer(wx.HORIZONTAL)
        message_sizer.Add(info_icon)
        message_sizer.AddSpacer((10, 10))
        message_sizer.Add(msg_text, flag=wx.EXPAND)

        vertical_sizer.Add(message_sizer, 1, wx.ALL, border=10)

        buttons_sizer = wx.BoxSizer(wx.HORIZONTAL)
        buttons_sizer.Add(ok_button)
        buttons_sizer.AddSpacer((5, -1))
        buttons_sizer.Add(cancel_button)

        vertical_sizer.Add(buttons_sizer, flag=wx.ALIGN_RIGHT | wx.ALL, border=10)

        panel.SetSizer(vertical_sizer)

        width, height = panel.GetBestSize()
        self.SetSize((width * 1.3, height * 1.3))

        self.Center()

        # Set up timer
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self._on_timer, self.timer)
        self.timer.Start(self.TIMER_INTERVAL)

    def _get_message(self):
        return self.message.format(self.timeout)

    # noinspection PyUnusedLocal
    def _on_timer(self, event):
        self.timeout -= 1
        self.msg_text.SetLabel(self._get_message())

        if self.timeout <= 0:
            self.EndModal(wx.ID_OK)

    def Destroy(self):
        self.timer.Stop()
        return super().Destroy()


# noinspection PyUnresolvedReferences
class LogGUI(wx.Frame):
    """Simple window for reading the STDERR.

    Attributes:
        FRAME_SIZE (tuple): Tuple that holds the frame size (width, height).

    Args:
        parent (MainFrame): Frame parent.

    """

    FRAME_SIZE: tuple[int, int] = (750, 200)

    def __init__(self, parent: MainFrame | OptionsFrame | None = None):
        super().__init__(parent, title=_("Log Viewer"), size=self.FRAME_SIZE)
        self.parent = parent
        self.app_icon: wx.Icon | None = self.parent.app_icon

        if self.app_icon:
            self.SetIcon(self.app_icon)

        self.panel = wx.Panel(self)

        self._text_area = wx.TextCtrl(
            self.panel, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.HSCROLL
        )

        sizer = wx.BoxSizer()
        sizer.Add(self._text_area, 1, wx.EXPAND)
        self.panel.SetSizerAndFit(sizer)

    def load(self, filename: str):
        """Load file content on the text area."""
        if Path(filename).exists():
            self._text_area.LoadFile(filename)



"""Contains test cases for the utils.py module."""

import locale
import sys
import unittest
from pathlib import Path
from unittest import mock

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui import utils


class TestUtils(unittest.TestCase):
    """Test case for utils functions"""

    def test_get_config_path(self):
        config_path: str = utils.get_config_path()
        self.assertIsNot(config_path, "")
        self.assertIsInstance(config_path, str)

    def test_decode_tuple(self):
        win_size: str = "740/490"
        win_size_tuple = utils.decode_tuple(win_size)
        self.assertEqual(len(win_size_tuple), 2)
        self.assertTrue(all(isinstance(size, int) for size in win_size_tuple))

    def test_encode_tuple(self):
        win_size_tuple = (740, 490)
        win_size: str = utils.encode_tuple(win_size_tuple)
        self.assertTrue("/" in win_size)

    @unittest.skipIf(
        not sys.platform.startswith("win"),
        "cp65001 encoding is only available on Windows",
    )
    @mock.patch("youtube_dl_gui.utils.locale_getpreferredencoding")
    def test_get_encoding(self, mock_getpreferredencoding):
        mock_getpreferredencoding.return_value = "cp65001"
        encoding = utils.get_encoding()
        self.assertEqual(encoding, "cp65001")
        mock_getpreferredencoding.assert_called_once()

    @mock.patch("youtube_dl_gui.utils.locale_getpreferredencoding")
    def test_get_encoding_error(self, mock_getpreferredencoding):
        mock_getpreferredencoding.side_effect = locale.Error()
        encoding = utils.get_encoding()
        self.assertEqual(encoding, "utf-8")
        mock_getpreferredencoding.assert_called_once()

    def test_get_key(self):
        dictionary = {"key": "value"}
        result = utils.get_key("value", dictionary)
        self.assertEqual(result, "key")
        result = utils.get_key("value2", dictionary)
        self.assertEqual(result, "")

    def test_get_time(self):
        timestamp = 1621991858.3169
        expected = {"seconds": 38, "minutes": 17, "hours": 1, "days": 18773}
        time_dict = utils.get_time(timestamp)
        self.assertDictEqual(time_dict, expected)


class TestToBytes(unittest.TestCase):
    """Test case for the to_bytes method."""

    def test_to_bytes_bytes(self):
        self.assertEqual(utils.to_bytes("596.00B"), 596.00)
        self.assertEqual(utils.to_bytes("133.55B"), 133.55)

    def test_to_bytes_kilobytes(self):
        self.assertEqual(utils.to_bytes("1.00KiB"), 1024.00)
        self.assertEqual(utils.to_bytes("5.55KiB"), 5683.20)

    def test_to_bytes_megabytes(self):
        self.assertEqual(utils.to_bytes("13.64MiB"), 14302576.64)
        self.assertEqual(utils.to_bytes("1.00MiB"), 1048576.00)

    def test_to_bytes_gigabytes(self):
        self.assertEqual(utils.to_bytes("1.00GiB"), 1073741824.00)
        self.assertEqual(utils.to_bytes("1.55GiB"), 1664299827.20)

    def test_to_bytes_terabytes(self):
        self.assertEqual(utils.to_bytes("1.00TiB"), 1099511627776.00)


class TestFormatBytes(unittest.TestCase):
    """Test case for the format_bytes method."""

    def test_format_bytes_bytes(self):
        self.assertEqual(utils.format_bytes(518.00), "518.00B")

    def test_format_bytes_kilobytes(self):
        self.assertEqual(utils.format_bytes(1024.00), "1.00KiB")

    def test_format_bytes_megabytes(self):
        self.assertEqual(utils.format_bytes(1048576.00), "1.00MiB")

    def test_format_bytes_gigabytes(self):
        self.assertEqual(utils.format_bytes(1073741824.00), "1.00GiB")

    def test_format_bytes_terabytes(self):
        self.assertEqual(utils.format_bytes(1099511627776.00), "1.00TiB")


class TestBuildCommand(unittest.TestCase):
    """Test case for the build_command method."""

    def setUp(self):
        self.url = "https://www.youtube.com/watch?v=aaaaaaaaaaa&list=AAAAAAAAAAA"

        self.options = ["-o", None, "-f", "mp4", "--ignore-config"]

        self.result = '{{ydl_bin}} -o "{{tmpl}}" -f mp4 --ignore-config "{url}"'.format(
            url=self.url
        )

    def run_tests(self, ydl_bin, tmpl):
        """Run the main test.

        Args:
            ydl_bin (str): Name of the youtube-dl binary
            tmpl (str): Youtube-dl output template

        """
        self.options[1] = tmpl  # Plug the template in our options

        result = self.result.format(ydl_bin=ydl_bin, tmpl=tmpl)

        self.assertEqual(utils.build_command(self.options, self.url, ydl_bin), result)

    def test_build_command_with_spaces_linux(self):
        tmpl = "/home/user/downloads/%(upload_date)s/%(id)s_%(playlist_id)s - %(format)s.%(ext)s"

        self.run_tests("youtube-dl", tmpl)

    def test_build_command_without_spaces_linux(self):
        tmpl = "/home/user/downloads/%(id)s.%(ext)s"

        self.run_tests("youtube-dl", tmpl)

    def test_build_command_with_spaces_windows(self):
        tmpl = "C:\\downloads\\%(upload_date)s\\%(id)s_%(playlist_id)s - %(format)s.%(ext)s"

        self.run_tests("youtube-dl.exe", tmpl)

    def test_build_command_without_spaces_windows(self):
        tmpl = "C:\\downloads\\%(id)s.%(ext)s"

        self.run_tests("youtube-dl.exe", tmpl)


class TestGetDefaultLang(unittest.TestCase):
    """Test case for the get_default_lang function."""

    @mock.patch("youtube_dl_gui.utils.locale_getdefaultlocale")
    def run_tests(self, ret_value, result, mock_getdefaultlocale):
        """Run the main test.

        Args:
            ret_value (tuple): Return tuple of the locale.getdefaultlocale module
            result (str): Result we want to see
            mock_getdefaultlocale (MagicMock): Mock object
        """
        mock_getdefaultlocale.return_value = ret_value
        lang = utils.get_default_lang()

        mock_getdefaultlocale.assert_called_once()
        self.assertEqual(lang, result)

    def test_get_default_lang(self):
        self.run_tests(("it_IT", "UTF-8"), "it_IT")

    def test_get_default_lang_none(self):
        self.run_tests((None, None), "en_US")

    def test_get_default_lang_empty(self):
        self.run_tests(("", ""), "en_US")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



# type: ignore[misc]
"""yt-dlg module that contains util functions.

Attributes:

    YOUTUBEDL_BIN (string): Youtube-dl binary filename.

"""
from __future__ import annotations

import locale
import math
import os
import subprocess
import sys
from pathlib import Path

from .info import __appname__

IS_WINDOWS = os.name == "nt"

YOUTUBEDL_BIN: str = "youtube-dl"
if IS_WINDOWS:
    YOUTUBEDL_BIN += ".exe"

YTDLP_BIN: str = "yt-dlp"
if IS_WINDOWS:
    YTDLP_BIN += ".exe"

FILESIZE_METRICS = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]

KILO_SIZE = 1024.0

locale_getdefaultlocale = locale.getdefaultlocale

locale_getpreferredencoding = locale.getpreferredencoding


def get_encoding() -> str:
    """Return system encoding, elsese utf-8"""
    try:
        encoding = locale_getpreferredencoding()
        _ = "TEST".encode(encoding)
    except locale.Error:
        encoding = "utf-8"

    return encoding


def startfile(file_path: str) -> None:
    if os.name == "nt":
        os.startfile(file_path)
    else:
        subprocess.call(("xdg-open", file_path))


os_startfile = startfile


def remove_file(filename: str) -> bool:
    file_path = Path(filename)
    if file_path.exists():
        file_path.unlink()
        return True

    return False


def remove_shortcuts(path: str) -> str:
    """Return given path after removing the shortcuts."""
    return path.replace("~", str(Path().home()))


def absolute_path(filename: str) -> str:
    """Return absolute path to the given file."""
    return str(Path(filename).resolve())


def open_file(file_path: str) -> bool:
    """Open file in file_path using the default OS application.

    Returns:
        True on success else False.

    """

    if not Path(file_path).exists():
        return False

    os_startfile(file_path)

    return True


def encode_tuple(tuple_to_encode: tuple[int, int]) -> str:
    """Turn size tuple into string."""
    return f"{tuple_to_encode[0]}/{tuple_to_encode[1]}"


def decode_tuple(encoded_tuple: str) -> tuple[int, int]:
    """Turn tuple string back to tuple."""
    s = encoded_tuple.split("/")
    return int(s[0]), int(s[1])


def check_path(path: str) -> None:
    """Create path if not exist."""
    if not Path(path).exists():
        os.makedirs(path)


# noinspection PyUnusedLocal
def get_config_path() -> str:
    """Return user config path.

    Note:
        Windows = %AppData% + app_name
        Linux   = ~/.config + app_name

    """
    ytdlg_path: str = ""

    if os.name == "nt":
        ytdlg_path = os.getenv("APPDATA", "")
    else:
        ytdlg_path = str(Path().home() / Path(".config"))

    return str(Path(ytdlg_path) / Path(__appname__.lower()))


# noinspection PyUnusedLocal
def shutdown_sys(password=None) -> bool:
    """Shuts down the system.
    Returns True if no errors occur else False.

    Args:
        password (string): SUDO password for linux.

    Note:
        On Linux you need to provide sudo password if you don't
        have elevated privileges.

    """
    info = None
    cmd = []
    encoding: str = get_encoding()

    kwargs = dict(
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        encoding=encoding,
        creationflags=0,
    )

    if os.name == "nt":
        cmd = ["shutdown", "/s", "/t", "1"]

        # Hide subprocess window
        info = subprocess.STARTUPINFO()
        info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        info.wShowWindow = subprocess.SW_HIDE

        kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
    else:
        kwargs["start_new_session"] = True

        if password:
            password = "%s\n" % password
            cmd = ["sudo", "-S", "/sbin/shutdown", "-h", "now"]
        else:
            cmd = ["/sbin/shutdown", "-h", "now"]

    shutdown_proc = subprocess.Popen(
        cmd, startupinfo=info, **kwargs
    )  # type: ignore[call-overload]

    output = shutdown_proc.communicate(password)[1]

    return not output or output == "Password:"


def get_time(seconds: float) -> dict[str, int]:
    """Convert given seconds to days, hours, minutes and seconds.

    Args:
        seconds (float): Time in seconds.

    Returns:
        Dictionary that contains the corresponding days, hours, minutes
        and seconds of the given seconds.

    """
    dtime = dict(seconds=0, minutes=0, hours=0, days=0)

    dtime["days"] = int(seconds / 86400)
    dtime["hours"] = int(seconds % 86400 / 3600)
    dtime["minutes"] = int(seconds % 86400 % 3600 / 60)
    dtime["seconds"] = int(seconds % 86400 % 3600 % 60)

    return dtime


def get_search_dirs(dir_name: str) -> list[Path]:
    return [
        Path(sys.argv[0]) / Path(dir_name),
        Path(__file__).parent / Path(dir_name),
    ]


# noinspection PyPep8Naming
def get_locale_file() -> str | None:
    """Search for yt_dlg locale file.

    Returns:
        The path to yt_dlg locale file if exists else None.

    Note:
        Paths that get_locale_file() func searches.

        __main__ dir, library dir

    """
    SEARCH_DIRS = get_search_dirs("locale")

    for directory in SEARCH_DIRS:
        if directory.is_dir():
            return str(directory)

    return None


# noinspection PyPep8Naming
def get_icon_file() -> str | None:
    """Search for yt_dlg app icon.

    Returns:
        The path to yt_dlg icon file if exists, else returns None.

    """
    pixmaps_dir = get_pixmaps_dir()

    if pixmaps_dir:
        ICON_NAME = "youtube-dl-gui.png"

        icon_file = Path(pixmaps_dir) / Path(ICON_NAME)

        if icon_file.exists():
            return str(icon_file)

    return None


# noinspection PyPep8Naming
def get_pixmaps_dir() -> str | None:
    """Return absolute path to the pixmaps icons folder.

    Note:
        Paths we search: __main__ dir, library dir

    """
    SEARCH_DIRS = get_search_dirs("data")

    for directory in SEARCH_DIRS:
        pixmaps_dir = directory / Path("pixmaps")

        if pixmaps_dir.is_dir():
            return str(pixmaps_dir)

    return None


def to_bytes(string: str) -> float:
    """Convert given youtube-dl size string to bytes."""
    value = 0.0
    index = 0

    for index, metric in enumerate(reversed(FILESIZE_METRICS)):
        if metric in string:
            value = float(string.split(metric)[0])
            break

    exponent = index * (-1) + (len(FILESIZE_METRICS) - 1)

    return round(value * (KILO_SIZE ** exponent), 2)


def format_bytes(bytes_: float) -> str:
    """Format bytes to youtube-dl size output strings."""
    exponent = 0 if bytes_ == 0.0 else int(math.log(bytes_, KILO_SIZE))
    suffix = FILESIZE_METRICS[exponent]
    output_value = bytes_ / (KILO_SIZE ** exponent)

    return f"{output_value:.2f}{suffix}"


def build_command(
    options_list: list[str], url: str, cli_backend: str = YOUTUBEDL_BIN
) -> str:
    """Build the CLI Backend command line string."""

    def escape(option: str) -> str:
        """Wrap option with double quotes if it contains special symbols."""

        special_symbols: list[str] = [" ", "(", ")"]

        for symbol in special_symbols:
            if symbol in option:
                return f'"{option}"'

        return option

    # If option has special symbols wrap it with double quotes
    # Probably not the best solution since if the option already contains
    # double quotes it will be a mess, see issue #173
    options: list[str] = [escape(option) for option in options_list]

    # Always wrap the url with double quotes
    url = f'"{url}"'

    return " ".join([cli_backend] + options + [url])


def get_default_lang() -> str:
    """Get default language using the 'locale' module."""
    default_lang, _ = locale_getdefaultlocale()

    return default_lang or "en_US"


def get_key(string: str, dictionary: dict[str, str], default: str = "") -> str:
    """Get key from a value in Dictionary. Return default if key doesn't exist"""
    for key, value in dictionary.items():
        if value == string:
            default = key
            return default
    return default



"""Contains test cases for the DownloadItem object."""

import sys
import unittest
from pathlib import Path

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.downloadmanager import DownloadItem

HOME = Path("/home/user")


class TestItemInit(unittest.TestCase):
    """Test case for DownloadItem init."""

    def test_init(self):
        url = "url"
        options = ["-f", "flv"]

        ditem = DownloadItem(url, options)

        self.assertEqual(ditem.stage, "Queued")
        self.assertEqual(ditem.url, url)
        self.assertEqual(ditem.options, options)
        self.assertEqual(ditem.object_id, hash(url + str(options)))

        self.assertEqual(ditem.path, "")
        self.assertEqual(ditem.filenames, [])
        self.assertEqual(ditem.extensions, [])
        self.assertEqual(ditem.filesizes, [])

        self.assertEqual(
            ditem.progress_stats,
            {
                "filename": url,
                "extension": "-",
                "filesize": "-",
                "percent": "0%",
                "speed": "-",
                "eta": "-",
                "status": "Queued",
                "playlist_size": "",
                "playlist_index": "",
            },
        )


class TestGetFiles(unittest.TestCase):
    """Test case for DownloadItem get_files method."""

    def setUp(self):
        self.ditem = DownloadItem("url", ["-f", "flv"])

    def test_get_files(self):
        path = HOME / Path("downloads")

        self.ditem.path = str(path)
        self.ditem.filenames = ["file1", "file2"]
        self.ditem.extensions = [".mp4", ".m4a"]

        self.assertEqual(
            self.ditem.get_files(),
            [
                str(path / Path("file1.mp4")),
                str(path / Path("file2.m4a")),
            ],
        )

    def test_get_files_no_data(self):
        self.assertEqual(self.ditem.get_files(), [])


class TestItemComparison(unittest.TestCase):
    """Test case for DownloadItem __eq__ method."""

    def test_equal_true(self):
        ditem1 = DownloadItem("url", ["-f", "flv"])
        ditem2 = DownloadItem("url", ["-f", "flv"])

        self.assertTrue(ditem1 == ditem2)

    def test_equal_false(self):
        ditem1 = DownloadItem("url", ["-f", "flv"])
        ditem2 = DownloadItem("url2", ["-f", "flv"])

        self.assertFalse(ditem1 == ditem2)

        ditem1 = DownloadItem("url", ["-f", "flv"])
        ditem2 = DownloadItem("url", ["-f", "mp4"])

        self.assertFalse(ditem1 == ditem2)


class TestSetItemStage(unittest.TestCase):
    """Test case for DownloadItem stage setter."""

    def setUp(self):
        self.ditem = DownloadItem("url", ["-f", "flv"])

    def test_set_stage_valid(self):
        self.ditem.stage = "Queued"
        self.assertEqual(self.ditem.stage, "Queued")
        self.assertEqual(self.ditem.progress_stats["status"], "Queued")

        self.ditem.stage = "Active"
        self.assertEqual(self.ditem.stage, "Active")
        self.assertEqual(self.ditem.progress_stats["status"], "Pre Processing")

        self.ditem.stage = "Completed"
        self.assertEqual(self.ditem.stage, "Completed")
        self.assertEqual(self.ditem.progress_stats["status"], "Finished")

        self.ditem.stage = "Paused"
        self.assertEqual(self.ditem.stage, "Paused")
        self.assertEqual(self.ditem.progress_stats["status"], "Paused")

        self.ditem.stage = "Error"
        self.assertEqual(self.ditem.stage, "Error")
        self.assertEqual(self.ditem.progress_stats["status"], "Error")

    def test_set_stage_invalid(self):
        raised = False

        try:
            self.ditem.stage = "some other status"
        except ValueError:
            raised = True

        self.assertTrue(raised)


class TestUpdateStats(unittest.TestCase):
    """Test case for DownloadItem update_stats method."""

    def setUp(self):
        self.ditem = DownloadItem("url", ["-f", "flv"])

    def test_update_stats(self):
        path = str(HOME)

        # 1st playlist item

        self.ditem.update_stats(
            {
                "filename": "somefilename.f1",
                "extension": ".mp4",
                "filesize": "9.45MiB",
                "percent": "2.0%",
                "speed": "200.00KiB/s",
                "eta": "00:38",
                "status": "Downloading",
                "path": path,
                "playlist_size": "10",
                "playlist_index": "1",
            }
        )

        self.assertEqual(self.ditem.path, path)
        self.assertEqual(self.ditem.filenames, ["somefilename.f1"])
        self.assertEqual(self.ditem.extensions, [".mp4"])

        # Do not update filesizes unless percentage is 100%
        # See https://github.com/MrS0m30n3/youtube-dl-gui/issues/162
        self.assertEqual(self.ditem.filesizes, [])

        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "somefilename.f1",
                "extension": ".mp4",
                "filesize": "9.45MiB",
                "percent": "2.0%",
                "speed": "200.00KiB/s",
                "eta": "00:38",
                "status": "Downloading",
                "playlist_size": "10",
                "playlist_index": "1",
            },
        )

        # Since the percentage is 100% this should update the filesizes list
        self.ditem.update_stats(
            {
                "filesize": "9.45MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
            }
        )

        self.assertEqual(self.ditem.filesizes, [9909043.20])

        self.ditem.update_stats(
            {
                "filename": "somefilename.f2",
                "extension": ".m4a",
                "filesize": "2.22MiB",
                "percent": "33.0%",
                "speed": "200.00KiB/s",
                "eta": "00:20",
                "status": "Downloading",
                "path": path,
            }
        )

        self.assertEqual(self.ditem.path, path)
        self.assertEqual(self.ditem.filenames, ["somefilename.f1", "somefilename.f2"])
        self.assertEqual(self.ditem.extensions, [".mp4", ".m4a"])
        self.assertEqual(self.ditem.filesizes, [9909043.20])

        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "somefilename.f2",
                "extension": ".m4a",
                "filesize": "2.22MiB",
                "percent": "33.0%",
                "speed": "200.00KiB/s",
                "eta": "00:20",
                "status": "Downloading",
                "playlist_size": "10",
                "playlist_index": "1",
            },
        )

        # Since the percentage is 100% this should update the filesizes list
        self.ditem.update_stats(
            {
                "filesize": "2.22MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
            }
        )

        self.assertEqual(self.ditem.filesizes, [9909043.20, 2327838.72])

        # Moving to the 2nd playlist item

        self.ditem.update_stats(
            {
                "filename": "someotherfilename.f1",
                "extension": ".mp4",
                "filesize": "10.25MiB",
                "percent": "50.0%",
                "speed": "200.00KiB/s",
                "eta": "00:38",
                "status": "Downloading",
                "path": path,
                "playlist_size": "10",
                "playlist_index": "2",
            }
        )

        # We must reset filenames, extensions & filesizes lists when changing playlist index
        # else the filesizes for the post processed files will be wrong
        self.assertEqual(self.ditem.filenames, ["someotherfilename.f1"])
        self.assertEqual(self.ditem.extensions, [".mp4"])
        self.assertEqual(self.ditem.filesizes, [])

        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "someotherfilename.f1",
                "extension": ".mp4",
                "filesize": "10.25MiB",
                "percent": "50.0%",
                "speed": "200.00KiB/s",
                "eta": "00:38",
                "status": "Downloading",
                "playlist_size": "10",
                "playlist_index": "2",
            },
        )

        # Since the percentage is 100% this should update the filesizes list
        self.ditem.update_stats(
            {
                "filesize": "10.25MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
            }
        )

        self.assertEqual(self.ditem.filesizes, [10747904.0])

        self.ditem.update_stats(
            {
                "filename": "someotherfilename.f2",
                "extension": ".m4a",
                "filesize": "3.33MiB",
                "percent": "33.0%",
                "speed": "200.00KiB/s",
                "eta": "00:30",
                "status": "Downloading",
                "path": path,
            }
        )

        self.assertEqual(self.ditem.path, path)
        self.assertEqual(
            self.ditem.filenames, ["someotherfilename.f1", "someotherfilename.f2"]
        )
        self.assertEqual(self.ditem.extensions, [".mp4", ".m4a"])
        self.assertEqual(self.ditem.filesizes, [10747904.0])

        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "someotherfilename.f2",
                "extension": ".m4a",
                "filesize": "3.33MiB",
                "percent": "33.0%",
                "speed": "200.00KiB/s",
                "eta": "00:30",
                "status": "Downloading",
                "playlist_size": "10",
                "playlist_index": "2",
            },
        )

        # Since the percentage is 100% this should update the filesizes list
        self.ditem.update_stats(
            {
                "filesize": "3.33MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
            }
        )

        self.assertEqual(self.ditem.filesizes, [10747904.0, 3491758.08])

        # Let's move to the 3rd playlist item

        # Here we mimic youtube-dl's "max downloads limit reached"
        # this line should not reset the filenames, extensions & filesizes lists
        # since we will lose the ability to play the last playlist item
        self.ditem.update_stats(
            {"status": "Downloading", "playlist_size": "10", "playlist_index": "3"}
        )

        self.assertEqual(
            self.ditem.filenames, ["someotherfilename.f1", "someotherfilename.f2"]
        )
        self.assertEqual(self.ditem.extensions, [".mp4", ".m4a"])
        self.assertEqual(self.ditem.filesizes, [10747904.0, 3491758.08])
        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "someotherfilename.f2",
                "extension": ".m4a",
                "filesize": "3.33MiB",
                "percent": "100%",
                "speed": "-",
                "eta": "-",
                "status": "Downloading",
                "playlist_size": "10",
                "playlist_index": "3",
            },
        )

    def test_update_stats_invalid_input(self):
        self.assertRaises(AssertionError, self.ditem.update_stats, [])

    def test_update_stats_empty_strings(self):
        self.ditem.update_stats(
            {
                "filename": "",
                "extension": "",
                "filesize": "",
                "percent": "",
                "speed": "",
                "eta": "",
                "status": "",
                "playlist_size": "",
                "playlist_index": "",
            }
        )

        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "url",
                "extension": "-",
                "filesize": "-",
                "percent": "0%",
                "speed": "-",
                "eta": "-",
                "status": "Queued",
                "playlist_size": "",
                "playlist_index": "",
            },
        )

    def test_update_stats_not_string(self):
        self.ditem.update_stats({"filename": None, "status": None, "eta": False})

        self.assertEqual(self.ditem.progress_stats["filename"], "url")
        self.assertEqual(self.ditem.progress_stats["status"], "Queued")
        self.assertEqual(self.ditem.progress_stats["eta"], "-")


class TestDownloadItemPrivate(unittest.TestCase):
    """Test case for private method of the DownloadItem."""

    def test_set_stage(self):
        ditem = DownloadItem("url", ["-f", "flv"])

        active_status = ["Pre Processing", "Downloading", "Post Processing"]
        complete_status = ["Finished", "Warning", "Already Downloaded"]
        error_status = ["Error", "Stopped", "Filesize Abort"]

        for status in active_status:
            ditem._set_stage(status)
            self.assertEqual(ditem.stage, "Active")

        for status in complete_status:
            ditem._set_stage(status)
            self.assertEqual(ditem.stage, "Completed")

        for status in error_status:
            ditem._set_stage(status)
            self.assertEqual(ditem.stage, "Error")

    def test_calc_post_proc_size(self):
        # REFACTOR Not an actual method
        # should transfer to TestUpdateStats
        ditem = DownloadItem("url", ["-f", "flv"])

        ditem.update_stats(
            {
                "filename": "file.f123",
                "extension": ".webm",
                "filesize": "10.00MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
                "path": "/home/user",
            }
        )

        ditem.update_stats(
            {
                "filename": "file.f456",
                "extension": ".m4a",
                "filesize": "3.45MiB",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Downloading",
                "path": "/home/user",
            }
        )

        # Mimic youtube-dl post process behaviour
        ditem.update_stats(
            {
                "filename": "file",
                "extension": ".webm",
                "percent": "100%",
                "speed": "",
                "eta": "",
                "status": "Post Processing",
            }
        )

        self.assertEqual(ditem.filesizes, [10485760.00, 3617587.20, 14103347.20])

        self.assertEqual(
            ditem.progress_stats,
            {
                "filename": "file",
                "extension": ".webm",
                "filesize": "13.45MiB",
                "percent": "100%",
                "speed": "-",
                "eta": "-",
                "status": "Post Processing",
                "playlist_size": "",
                "playlist_index": "",
            },
        )


class TestReset(unittest.TestCase):
    """Test case for the DownloadItem reset method."""

    def setUp(self):
        self.ditem = DownloadItem("url", ["-f", "flv"])

    def test_reset_completed_stage(self):
        self.ditem._stage = "Completed"
        self.ditem.path = str(HOME)
        self.ditem.filenames = ["file"]
        self.ditem.extensions = [".mp4"]
        self.ditem.filesizes = [123456.00]
        self.ditem.progress_stats = {
            "filename": "file",
            "extension": ".mp4",
            "filsize": "6.66MiB",
            "percent": "100%",
            "speed": "-",
            "eta": "-",
            "status": "Finished",
            "playlist_size": "",
            "playlist_index": "",
        }

        self.ditem.reset()

        self.assertEqual(self.ditem._stage, "Queued")
        self.assertEqual(self.ditem.path, "")
        self.assertEqual(self.ditem.filenames, [])
        self.assertEqual(self.ditem.extensions, [])
        self.assertEqual(self.ditem.filesizes, [])
        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "url",
                "extension": "-",
                "filesize": "-",
                "percent": "0%",
                "speed": "-",
                "eta": "-",
                "status": "Queued",
                "playlist_size": "",
                "playlist_index": "",
            },
        )

    def test_reset_error_stage(self):
        self.ditem._stage = "Error"
        self.ditem.path = str(HOME)
        self.ditem.filenames = ["file1", "file2", "file"]
        self.ditem.extensions = [".mp4", ".m4a", ".mp4"]
        self.ditem.filesizes = [1234.00, 3421.00, 4655.00]
        self.ditem.progress_stats = {
            "filename": "file",
            "extension": ".mp4",
            "filsize": "9.45MiB",
            "percent": "100%",
            "speed": "-",
            "eta": "-",
            "status": "Error",
            "playlist_size": "10",
            "playlist_index": "8",
        }

        self.ditem.reset()

        self.assertEqual(self.ditem._stage, "Queued")
        self.assertEqual(self.ditem.path, "")
        self.assertEqual(self.ditem.filenames, [])
        self.assertEqual(self.ditem.extensions, [])
        self.assertEqual(self.ditem.filesizes, [])
        self.assertEqual(
            self.ditem.progress_stats,
            {
                "filename": "url",
                "extension": "-",
                "filesize": "-",
                "percent": "0%",
                "speed": "-",
                "eta": "-",
                "status": "Queued",
                "playlist_size": "",
                "playlist_index": "",
            },
        )

    def test_reset_paused_stage(self):
        self.ditem._stage = "Paused"
        # No need to change filanames, extension, etc
        # since everything in pause state has the default value

        self.ditem.reset()
        self.assertEqual(self.ditem._stage, "Queued")

    def test_reset_active_stage(self):
        self.ditem._stage = "Active"
        self.ditem.path = str(HOME)
        self.ditem.filenames = ["file1"]
        self.ditem.extensions = [".mp4"]
        self.ditem.filesizes = []
        self.ditem.progress_stats = {
            "filename": "file1",
            "extension": ".mp4",
            "filsize": "9.45MiB",
            "percent": "75.5%",
            "speed": "200.00KiB/s",
            "eta": "00:10",
            "status": "Downloading",
        }

        self.assertRaises(RuntimeError, self.ditem.reset)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



"""Contains test cases for the parsers module."""

import sys
import unittest
from pathlib import Path

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.parsers import OptionsParser

SAVE_PATH: str = str(Path("/home/user/Workplace/test/youtube/%(title)s.%(ext)s"))


class TestParse(unittest.TestCase):
    """Test case for OptionsParser parse method."""

    def setUp(self):
        # Create the options_dict based on the OptionHolder
        # items inside the OptionsParser object
        self.options_dict = {
            item.name: item.default_value for item in OptionsParser()._ydl_options
        }

        # Add extra options used by the OptionsParser.parse method
        self.options_dict["save_path"] = "/home/user/Workplace/test/youtube/"
        self.options_dict["cmd_args"] = ""
        self.options_dict["output_format"] = "1"  # Title
        self.options_dict["second_video_format"] = "0"
        self.options_dict["min_filesize_unit"] = ""
        self.options_dict["max_filesize_unit"] = ""

    def check_options_parse(self, expected_options):
        options_parser = OptionsParser()

        self.assertListEqual(
            sorted(options_parser.parse(self.options_dict)), sorted(expected_options)
        )

    def test_parse_to_audio_requirement_bug(self):
        """Test case for the 'to_audio' requirement."""

        self.options_dict["audio_quality"] = "9"
        self.options_dict["audio_format"] = "mp3"
        self.options_dict["embed_thumbnail"] = True

        expected_cmd_list = [
            "--newline",
            "-x",
            "--audio-format",
            "mp3",
            "--embed-thumbnail",
            "--audio-quality",
            "9",
            "-o",
            SAVE_PATH,
        ]

        self.check_options_parse(expected_cmd_list)

        # Setting 'to_audio' to True should return the same results
        # since the '-x' flag is already set on audio extraction
        self.options_dict["to_audio"] = True
        self.check_options_parse(expected_cmd_list)

        # Setting 'to_audio' to True without audio_format
        self.options_dict["audio_quality"] = "3"
        self.options_dict["audio_format"] = ""
        self.options_dict["embed_thumbnail"] = False

        expected_cmd_list = [
            "--newline",
            "-x",
            "--audio-quality",
            "3",
            "-o",
            SAVE_PATH,
        ]

        self.check_options_parse(expected_cmd_list)

        # Setting 'to_audio' to True with default (mid) audio_quality
        self.options_dict["audio_quality"] = "5"
        self.options_dict["audio_format"] = ""
        self.options_dict["embed_thumbnail"] = True

        expected_cmd_list = [
            "--newline",
            "-x",
            "--embed-thumbnail",
            "--audio-quality",
            "5",
            "-o",
            SAVE_PATH,
        ]

        self.check_options_parse(expected_cmd_list)

    def test_parse_cmd_args_with_quotes(self):
        """Test the youtube-dl cmd line args parsing when quotes are presented.

        See: https://github.com/MrS0m30n3/youtube-dl-gui/issues/54

        """

        self.options_dict["video_format"] = "mp4"

        # Test with three quoted 'cmd_args'
        self.options_dict[
            "cmd_args"
        ] = '--recode-video mkv --postprocessor-args "-codec copy -report"'

        expected_cmd_list = [
            "--newline",
            "-f",
            "mp4",
            "-o",
            SAVE_PATH,
            "--recode-video",
            "mkv",
            "--postprocessor-args",
            "-codec copy -report",
        ]

        self.check_options_parse(expected_cmd_list)

        # Test with two quoted 'cmd_args'
        self.options_dict["cmd_args"] = '--postprocessor-args "-y -report"'

        expected_cmd_list = [
            "--newline",
            "-f",
            "mp4",
            "-o",
            SAVE_PATH,
            "--postprocessor-args",
            "-y -report",
        ]

        self.check_options_parse(expected_cmd_list)

        # Test with one quoted 'cmd_arg' followed by other cmd line args
        self.options_dict["cmd_args"] = '--postprocessor-args "-y" -v'

        expected_cmd_list = [
            "--newline",
            "-f",
            "mp4",
            "-o",
            SAVE_PATH,
            "--postprocessor-args",
            "-y",
            "-v",
        ]

        self.check_options_parse(expected_cmd_list)

        # Test the example presented in issue #54
        self.options_dict["cmd_args"] = '-f "(mp4)[width<1300]"'
        self.options_dict["video_format"] = "0"  # Set video format to 'default'

        expected_cmd_list = [
            "--newline",
            "-o",
            SAVE_PATH,
            "-f",
            "(mp4)[width<1300]",
        ]

        self.check_options_parse(expected_cmd_list)

        # Test for single quotes
        self.options_dict["cmd_args"] = "-f '(mp4)[width<1300]'"

        expected_cmd_list = [
            "--newline",
            "-o",
            SAVE_PATH,
            "-f",
            "(mp4)[width<1300]",
        ]

        self.check_options_parse(expected_cmd_list)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



# type: ignore[misc]
"""yt-dlg wx Application

Example:
    In order to load the GUI from a python script.

        from youtube_dl_gui import app

        app.main()

"""
from __future__ import annotations

import sys
from pathlib import Path
from typing import Any, Callable

import wx

from .info import __appname__, __packagename__
from .logmanager import LogManager
from .mainframe import MainFrame
from .optionsmanager import OptionsManager
from .utils import YOUTUBEDL_BIN, get_config_path, get_locale_file
from .version import __version__

_: Callable[[str], str] = wx.GetTranslation

# Set config path and create options and log managers
config_path: str = get_config_path()

opt_manager = OptionsManager(config_path)
log_manager = None

if opt_manager.options.get("enable_log", True):
    log_manager = LogManager(config_path, opt_manager.options.get("log_time", True))

locale_dir: str = get_locale_file() or "."


# noinspection PyPep8Naming
def _displayHook(obj: Any) -> None:
    """
    Install a custom displayhook to keep Python from setting the global
    _ (underscore) to the value of the last evaluated expression.  If
    we don't do this, our mapping of _ to gettext can get overwritten.
    This is useful/needed in interactive debugging with PyShell.

    :param Any `obj`: object to string representation

    """
    if obj is not None:
        print(repr(obj))


# noinspection PyPep8Naming,PyAttributeOutsideInit
class BaseApp(wx.App):
    """Base wx Application"""

    def OnInit(self) -> bool:
        super().OnInit()
        sys.displayhook = _displayHook

        self.appName: str = __appname__
        self.locale: wx.Locale | None = None
        wx.Locale.AddCatalogLookupPathPrefix(locale_dir)
        self.updateLanguage(opt_manager.options.get("locale_name", "en_US"))

        return True

    def updateLanguage(self, lang: str) -> None:
        """
        Update the language to the requested one.

        Make *sure* any existing locale is deleted before the new
        one is created.  The old C++ object needs to be deleted
        before the new one is created, and if we just assign a new
        instance to the old Python variable, the old C++ locale will
        not be destroyed soon enough, likely causing a crash.

        Args:
             lang (str): one of the supported language codes

        """

        # Supported Languages
        # Lang code = <ISO 639-1>_<ISO 3166-1 alpha-2>
        supLang: dict[str, int] = {
            "ar_SA": wx.LANGUAGE_ARABIC,
            "cs_CZ": wx.LANGUAGE_CZECH,
            "de_DE": wx.LANGUAGE_GERMAN,
            "en_US": wx.LANGUAGE_ENGLISH_US,
            "fr_FR": wx.LANGUAGE_FRENCH,
            "es_CU": wx.LANGUAGE_SPANISH,
            "it_IT": wx.LANGUAGE_ITALIAN,
            "ja_JP": wx.LANGUAGE_JAPANESE,
            "ko_KR": wx.LANGUAGE_KOREAN,
            "pl_PL": wx.LANGUAGE_POLISH,
            "pt_BR": wx.LANGUAGE_PORTUGUESE_BRAZILIAN,
            "ru_RU": wx.LANGUAGE_RUSSIAN,
            "es_ES": wx.LANGUAGE_SPANISH,
            "sq_AL": wx.LANGUAGE_ALBANIAN,
            "sk_SK": wx.LANGUAGE_SLOVAK,
            "zh_CN": wx.LANGUAGE_CHINESE_SIMPLIFIED,
            "zh_TW": wx.LANGUAGE_CHINESE_TRADITIONAL,
        }

        selLang: int = supLang.get(lang, wx.LANGUAGE_ENGLISH_US)

        if self.locale:
            assert sys.getrefcount(self.locale) <= 2
            del self.locale

        # create a locale object for this language
        self.locale = wx.Locale(selLang)

        if self.locale.IsOk():
            self.locale.AddCatalog(__packagename__)
        else:
            self.locale = None


# BaseApp and MainFrame
app = BaseApp(redirect=False)
frame = MainFrame(opt_manager, log_manager)


def main() -> int:
    """
    The real main. Calls the main app (`BaseApp`) windows.

    Print the version if pass option -v|--version
    """
    _error = 0

    if len(sys.argv) > 1 and sys.argv[1].lower() in ["-v", "--version"]:
        print(f"{__appname__} {__version__}")
        return _error

    youtubedl_path: Path = (
        Path(opt_manager.options.get("youtubedl_path", ".")) / YOUTUBEDL_BIN
    )

    frame.Show()

    if opt_manager.options.get("disable_update", False) and not youtubedl_path.exists():
        wx.MessageBox(
            _("Failed to locate CLI Backend and updates are disabled"),
            _("Error"),
            wx.OK | wx.ICON_ERROR,
        )
        _error = 1
        frame.close()

    app.MainLoop()

    return _error



"""Contains test cases for the widgets.py module."""

import sys
import unittest
from pathlib import Path
from unittest import mock

import wx

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.widgets import ListBoxWithHeaders


class TestListBoxWithHeaders(unittest.TestCase):
    """Test cases for the ListBoxWithHeaders widget."""

    def setUp(self):
        self.app = wx.App()

        self.frame = wx.Frame(None)
        self.listbox = ListBoxWithHeaders(self.frame)

        self.listbox.add_header("Header")
        self.listbox.add_items(["item%s" % i for i in range(10)])

    def tearDown(self):
        self.frame.Destroy()
        del self.app

    def test_find_string_header_found(self):
        self.assertEqual(
            self.listbox.FindString(
                "Header",
            ),
            0,
        )

    def test_find_string_header_not_found(self):
        self.assertEqual(
            self.listbox.FindString(
                "Header2",
            ),
            wx.NOT_FOUND,
        )

    def test_find_string_item_found(self):
        self.assertEqual(
            self.listbox.FindString(
                "item1",
            ),
            2,
        )

    def test_find_string_item_not_found(self):
        self.assertEqual(
            self.listbox.FindString(
                "item",
            ),
            wx.NOT_FOUND,
        )

    def test_get_string_header(self):
        self.assertEqual(self.listbox.GetString(0), "Header")

    def test_get_string_item(self):
        self.assertEqual(self.listbox.GetString(10), "item9")

    def test_get_string_item_not_found(self):
        self.assertEqual(self.listbox.GetString(11), "")

    def test_get_string_item_negative_index(self):
        self.assertEqual(self.listbox.GetString(-1), "")

    def test_insert_items(self):
        self.listbox.SetSelection(1)

        self.listbox.InsertItems(["new_item1", "new_item2"], 1)
        self.assertEqual(self.listbox.GetString(1), "new_item1")
        self.assertEqual(self.listbox.GetString(2), "new_item2")
        self.assertEqual(self.listbox.GetString(3), "item0")

        self.assertTrue(self.listbox.IsSelected(3))  # Old selection + 2

    def test_set_selection_header(self):
        self.listbox.SetSelection(0)
        self.assertFalse(self.listbox.IsSelected(0))

    def test_set_selection_item_valid_index(self):
        self.listbox.SetSelection(1)
        self.assertEqual(self.listbox.GetSelection(), 1)

    def test_set_selection_item_invalid_index(self):
        self.listbox.SetSelection(1)
        self.assertEqual(self.listbox.GetSelection(), 1)

        self.listbox.SetSelection(wx.NOT_FOUND)
        self.assertEqual(self.listbox.GetSelection(), wx.NOT_FOUND)

    def test_set_string_item(self):
        self.listbox.SetString(1, "item_mod0")
        self.assertEqual(self.listbox.GetString(1), "item_mod0")

    def test_set_string_header(self):
        self.listbox.SetString(0, "New header")
        self.assertEqual(self.listbox.GetString(0), "New header")

        # Make sure that the header is not selectable
        self.listbox.SetSelection(0)
        self.assertFalse(self.listbox.IsSelected(0))

    def test_set_string_selection_header(self):
        self.assertFalse(self.listbox.SetStringSelection("Header"))
        self.assertFalse(self.listbox.IsSelected(0))

    def test_set_string_selection_item(self):
        self.assertTrue(self.listbox.SetStringSelection("item1"))
        self.assertTrue(self.listbox.IsSelected(2))

    def test_get_string_selection(self):
        self.listbox.SetSelection(1)
        self.assertEqual(self.listbox.GetStringSelection(), "item0")

    def test_get_string_selection_empty(self):
        self.assertEqual(self.listbox.GetStringSelection(), "")

    # wx.ItemContainer methods

    def test_append(self):
        self.listbox.Append("item666")
        self.assertEqual(self.listbox.GetString(11), "item666")

    def test_append_items(self):
        self.listbox.AppendItems(["new_item1", "new_item2"])
        self.assertEqual(self.listbox.GetString(11), "new_item1")
        self.assertEqual(self.listbox.GetString(12), "new_item2")

    def test_clear(self):
        self.listbox.Clear()
        self.assertEqual(self.listbox.GetItems(), [])

    def test_delete(self):
        self.listbox.Delete(0)
        self.assertEqual(self.listbox.GetString(0), "item0")

        # Test item selection
        self.listbox.SetSelection(0)
        self.assertTrue(self.listbox.IsSelected(0))

    # Test object extra methods

    def test_add_header(self):
        self.listbox.add_header("Header2")
        self.listbox.SetSelection(11)
        self.assertFalse(self.listbox.IsSelected(11))

    @mock.patch("wx.ListBox.Append")
    def test_add_item_with_prefix(self, mock_append):
        self.listbox.add_item("new_item")
        mock_append.assert_called_once_with(
            ListBoxWithHeaders.TEXT_PREFIX + "new_item", None
        )

    @mock.patch("wx.ListBox.Append")
    def test_add_item_without_prefix(self, mock_append):
        self.listbox.add_item("new_item", with_prefix=False)
        mock_append.assert_called_once_with("new_item", None)

    @mock.patch("wx.ListBox.AppendItems")
    def test_add_items_with_prefix(self, mock_append):
        self.listbox.AppendItems(["new_item1", "new_item2"])
        mock_append.assert_called_once_with(
            [
                ListBoxWithHeaders.TEXT_PREFIX + "new_item1",
                ListBoxWithHeaders.TEXT_PREFIX + "new_item2",
            ]
        )

    @mock.patch("wx.ListBox.AppendItems")
    def test_add_items_without_prefix(self, mock_append):
        self.listbox.AppendItems(["new_item1", "new_item2"], with_prefix=False)
        mock_append.assert_called_once_with(["new_item1", "new_item2"])


def main():
    unittest.main()


if __name__ == "__main__":
    main()



"""Contains test cases for the updatemanager.py module."""

import sys
import unittest
from io import StringIO
from pathlib import Path
from unittest import mock

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.updatemanager import UpdateThread
from youtube_dl_gui.utils import YOUTUBEDL_BIN

DATA_JSON = PATH.joinpath("fixtures/updatemanager_releases_latest.json").read_text(
    encoding="utf-8"
)


class TestUpdateThread(unittest.TestCase):
    @mock.patch("youtube_dl_gui.downloadmanager.UpdateThread", autospec=True)
    def test_init_check_sig(self, mock_update):
        update_thread = mock_update(opt_manager=None, quiet=False, daemon=False)
        self.assertIsInstance(update_thread, UpdateThread)

    @mock.patch("youtube_dl_gui.updatemanager.open")
    @mock.patch("youtube_dl_gui.updatemanager.check_path")
    @mock.patch("youtube_dl_gui.updatemanager.urlopen")
    @mock.patch("youtube_dl_gui.optionsmanager.OptionsManager")
    def test_downloadmanager(
        self, mock_opt_manager, mock_urlopen, mock_check_path, mock_open
    ):
        config_path = "/home/user/.config"
        opt_manager = mock_opt_manager(config_path)
        opt_manager.options = {
            "youtubedl_path": config_path,
            "cli_backend": YOUTUBEDL_BIN,
        }

        mock_urlopen.side_effect = [StringIO(DATA_JSON), StringIO("")]

        update_thread = UpdateThread(opt_manager)
        update_thread.join()

        mock_check_path.assert_called_once()
        self.assertEqual(update_thread.cli_backend, YOUTUBEDL_BIN)
        self.assertEqual(update_thread.name, "UpdateManager")


def main():
    unittest.main()


if __name__ == "__main__":
    main()
